#!/bin/env python
"""
Module for outputting data in wiggle format.

See the README.txt file (part of this source code distribution) for usage,
contact info, how to cite, etc.

Copyright 2010 Andrew V. Uzilov and The Regents of the University of California
All Rights Reserved

This file is part of FragSeq_v0.0.1.

FragSeq_v0.0.1 is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FragSeq_v0.0.1 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FragSeq_v0.0.1.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import sys


class WiggleWriter():
    def __init__ (self, fileName,
                  name=None, description=None, visibility=None, color=None,
                  maxHeightPixels=None, viewLimits=None, extra=None):
        self.out = open (fileName, mode='w')
        headerLines = ['track type=wiggle_0']
        if name is not None:
            assert len (name) <= 15, \
                   'Track label (table name) "%s" too long: %d chars (must be <= 15 chars).' \
                   % (name, len (name))
            headerLines.append ('name="%s"' % name)
        if description is not None:
            headerLines.append ('description="%s"' % description)
        if visibility is not None:
            assert visibility == 'full' or visibility == 'dense' \
                or visibility == 'hide', \
                "Invalid 'visiblity' setting: '%s'." % visibility
            headerLines.append ('visibility=%s' % visibility)
        if color is not None:
            assert re.compile ('^\d+,\d+,\d+$').search (color) is not None
            headerLines.append ('color=%s' % color)
        if maxHeightPixels is not None:
            assert len (maxHeightPixels) == 3
            hMax     = int (maxHeightPixels[0])
            hDefault = int (maxHeightPixels[1])
            hMin     = int (maxHeightPixels[2])
            assert hMin > 0
            assert hMax > 0
            assert hDefault > 0
            assert hMin <= hMax
            assert hDefault >= hMin and hDefault <= hMax
            headerLines.append (
                'maxHeightPixels=%d:%d:%d' % (hMax, hDefault, hMin))
            
        if viewLimits is not None:
            assert len (viewLimits) == 2
            minLim = float (viewLimits[0])
            maxLim = float (viewLimits[1])
            headerLines.append ('viewLimits=%f:%f' % (minLim, maxLim))
            # For some reason, the following is necessary to make browser
            # actually use viewLimits settings, even though it ought to be
            # the default.
            headerLines.append ('autoScale=off')

        if extra is not None:
            headerLines.append (extra)
        
        self.out.write (' '.join (headerLines) + '\n')

 
    def writeStanza (self, stanzaFormat, **moreNamedArgs):
        """Write a wiggle stanza for a contiguous sequence of values.

        Caller can specify to use only a subrange of values.

        'threshold' means write scores ABOVE threshold only (not inclusive).

        Only 'fixedStep'-type stanzas work right now.
        """
        #! General options.
        if moreNamedArgs.has_key ('threshold'):
            threshold = float (moreNamedArgs['threshold'])
        else:
            threshold = float ('-inf')  # effectively no threshold
        if moreNamedArgs.has_key ('filterVals'):
            assert isinstance (moreNamedArgs['filterVals'], set)
        else:
            moreNamedArgs['filterVals'] = set()  # effectively no filter vals

        #! Options specific to a certain wiggle type.
        if stanzaFormat == 'bed':
            assert False, "Not yet implemented."
            
        elif stanzaFormat == 'variableStep':
            assert False, "Not yet implemented."
            
        elif stanzaFormat == 'fixedStep':
            # Get parameters.
            for argName in ['chrom', 'start', 'values']:  # required args
                assert moreNamedArgs.has_key (argName)
            start = int (moreNamedArgs['start']) + 1  # convert to 1-based
            assert start >= 1
            step, span = 1, 1  # defaults
            if moreNamedArgs.has_key ('step'):
                step = int (moreNamedArgs['step'])
                assert step >= 1
            if moreNamedArgs.has_key ('span'):
                span = int (moreNamedArgs['span'])
                assert span >= 1

            # How to iterate over values.
            if moreNamedArgs.has_key ('strand'):
                # Strand specified expicitly.
                if moreNamedArgs['strand'] == '+':
                    vStep = 1
                else:
                    assert moreNamedArgs['strand'] == '-'
                    vStep = -1
            else:
                # Assume plus strand (iterate from L to R).
                vStep = 1
            if moreNamedArgs.has_key ('vStart') or \
               moreNamedArgs.has_key ('vEnd'):
                assert moreNamedArgs.has_key ('vStart')
                assert moreNamedArgs.has_key ('vEnd')
                vStart = int (moreNamedArgs['vStart'])
                vEnd   = int (moreNamedArgs['vEnd'])
                assert vStart >= 0
                assert vEnd   <= len (moreNamedArgs ['values'])
                assert vStart < vEnd
            else:
                vStart = 0
                vEnd   = len (moreNamedArgs ['values'])
            if vStep == -1:
                # Flip for passing to xrange.
                vStart, vEnd = vEnd - 1, vStart - 1
            
            # Write values to file.
            valStack = []
            for i in xrange (vStart, vEnd, vStep):
                val = moreNamedArgs['values'][i]
                try:
                    if val is not None:
                        val = float (val)
                except TypeError:
                    print >> sys.stderr, \
                        'Value "%s" at index %d of values passed to writeStanza() is neither None, nor can be parsed as float.' \
                        % (moreNamedArgs['values'][i], i)
                    sys.exit (1)
                
                if val is not None and val > threshold \
                       and val not in moreNamedArgs['filterVals']:
                    if len (valStack) == 0:
                        # Hit first non-None value above threshold
                        # (start of run of above threshold values).
                        stackStart = start
                    valStack.append (val)
                else:
                    # Hit a value not meeting threshold or None (missing data)
                    # Dump stack contents (all non-None and above threshold).
                    if len (valStack) > 0:
                        self.out.write (
                            'fixedStep chrom=%s start=%d step=%d span=%d\n'
                            % (moreNamedArgs['chrom'], stackStart,
                               step, span))
                        for val in valStack:
                            self.out.write ('%f\n' % val)
                        valStack = []  # reset stack
                start += step
                
            if len (valStack) > 0:
                # Still some stuff left in the stack.
                self.out.write (
                    'fixedStep chrom=%s start=%d step=%d span=%d\n'
                    % (moreNamedArgs['chrom'], stackStart,
                       step, span))
                for val in valStack:
                    self.out.write ('%f\n' % val)
